from typing import Optional, Callable, List

import numpy as np

from torusgrid.fields import RealField2D
from torusgrid import field_util as fu
from torusgrid.dynamics import EvolverCallBack, FreeEnergyFunctional2D

from ..core.base import PFCFreeEnergyFunctional
from ..core.evolution import PFCMinimizer
from ..pfc import PFC



def find_coexistent_mu(solid_field: RealField2D, dt: float, eps: float, mu_min: float, mu_max: float, 
        relax=True, max_iters: int=None, precision: float=0., **minimizer_kwargs) -> dict:

    if mu_min >= mu_max:
        raise ValueError('mu_min must be smaller than mu_max')
    
    if max_iters is None and precision == 0.:
        raise ValueError('binary search will not stop with max_iters=None and precision=0') 

    fef = PFCFreeEnergyFunctional(eps)
    model_s = PFC(solid_field)
    model_l = PFC(fu.liquefy(solid_field))

    solid_psi = model_s.field.psi.copy()
    liquid_psi = model_l.field.psi.copy()
    
    rec = {'mu': [], 'omega_s': [], 'omega_l': [], 'mu_min_initial': mu_min, 'mu_max_initial': mu_max}
    
    print('===========================================================================')

    for i in range(max_iters):
        if mu_max - mu_min <= precision:
            break
        mu = (mu_min + mu_max) / 2
        print(f'current bounds: {mu_min} ~ {mu_max}')

        if relax:
            model_s.evolve(minimizer='relax', dt=dt, eps=eps, mu=mu, expansion_rate=1., **minimizer_kwargs)
            model_l.field.set_size(model_s.field.Lx, model_s.field.Ly)

        model_s.evolve(minimizer='mu', dt=dt, eps=eps, mu=mu, **minimizer_kwargs)
        omega_s = fef.mean_grand_potential_density(model_s.field, mu)

        if is_liquid(model_s.field.psi, tol=1e-4):
            print(f'solid field was liquefied during minimization with mu={mu}')
            break
            #mu_max = mu
            #model_s.field.set_psi(solid_psi)
            #continue

        model_l.evolve(minimizer='mu', dt=0.001, eps=eps, mu=mu, **minimizer_kwargs)
        omega_l = fef.mean_grand_potential_density(model_l.field, mu)

        rec['mu'].append(mu)
        rec['omega_s'].append(omega_s)
        rec['omega_l'].append(omega_l)

        if omega_s < omega_l:
            mu_min = mu

        if omega_s > omega_l:
            mu_max = mu


        print('------------------------------------------------------------------------')

    rec['mu_min_final'] = mu_min
    rec['mu_max_final'] = mu_max
    return rec


# works for a generic free energy functional
# fef: free energy functional
def find_coexistent_mu_general(
    solid_field: RealField2D,
    mu_min: float, mu_max: float,
    fef: FreeEnergyFunctional2D,
    minimizer_supplier: Callable[[RealField2D, float], PFCMinimizer], 
    relaxer_supplier: Optional[Callable[[RealField2D, float], PFCMinimizer]]=None,
    max_iters: Optional[int]=None, precision: float=0.,
    relax_callbacks: List[EvolverCallBack]=[],
    minim_callbacks: List[EvolverCallBack]=[],
    **minimizer_kwargs):

    if mu_min >= mu_max:
        raise ValueError('mu_min must be smaller than mu_max')
    
    if max_iters is None and precision == 0.:
        raise ValueError('binary search will not stop with max_iters=None and precision=0') 

    sol = solid_field
    liq = fu.liquefy(solid_field)

    #solid_psi = model_s.field.psi.copy()
    #liquid_psi = model_l.field.psi.copy()
    
    rec = {'mu': [], 'omega_s': [], 'omega_l': [], 'mu_min_initial': mu_min, 'mu_max_initial': mu_max}
    
    print('===========================================================================')

    for i in range(max_iters):
        if mu_max - mu_min <= precision:
            break
        mu = (mu_min + mu_max) / 2
        print(f'current bounds: {mu_min} ~ {mu_max}')

        if relaxer_supplier:
            # relax solid and resize liquid accordingly
            relaxer = relaxer_supplier(sol, mu)
            relaxer.run_nonstop(31, callbacks=relax_callbacks, **minimizer_kwargs)
            liq.set_size(sol.Lx, sol.Ly)

        # evolve solid under constant mu
        minim = minimizer_supplier(sol, mu)
        minim.run_nonstop(31, callbacks=minim_callbacks, **minimizer_kwargs)

        # calculate solid mean grand potential
        omega_s = fef.mean_grand_potential_density(sol, mu)

        if is_liquid(sol.psi, tol=1e-4):
            print(f'solid field was liquefied during minimization with mu={mu}')
            break

        # evolve liquid under constant mu
        minim = minimizer_supplier(liq, mu)
        minim.run_nonstop(31, callbacks=minim_callbacks, **minimizer_kwargs)
        # calculate liquid mean grand potential
        omega_l = fef.mean_grand_potential_density(liq, mu)

        rec['mu'].append(mu)
        rec['omega_s'].append(omega_s)
        rec['omega_l'].append(omega_l)

        if omega_s < omega_l:
            mu_min = mu

        if omega_s > omega_l:
            mu_max = mu


        print('------------------------------------------------------------------------')

    rec['mu_min_final'] = mu_min
    rec['mu_max_final'] = mu_max
    return rec


def is_liquid(psi: np.ndarray, tol=1e-5):
    return np.max(psi) - np.min(psi) <= tol


# [interface] -> [left][right]
# elongation: [delta_liq][------left------][delta_sol][delta_sol][------right------][delta_liq]
def evolve_and_elongate_interface(
        ifc: RealField2D, delta_sol: RealField2D, delta_liq: RealField2D,
        minimizer_supplier: Callable[[PFC], PFCMinimizer]=None,
        minimizer: str=None, dt: float=None, eps: float=None, mu: float=None, 
        N_steps: int=31, N_epochs: Optional[int]=None, interrupt_handler: Optional[Callable]=None,
        display_format: Optional[str]=None, callbacks: List[Callable]=[],
        verbose=False):

    try:
        assert delta_sol.Ny == delta_liq.Ny == ifc.Ny
        assert delta_sol.Ly == delta_liq.Ly == ifc.Ly
    except AssertionError:
        raise ValueError('delta_sol, delta_liq, and ifc must have the same Y dimensions')


    if verbose:
        print(f'Lx={ifc.Lx}, Ly={ifc.Ly}')
        print(f'evolving interface')
    model = PFC(ifc)

    model.evolve(
            minimizer_supplier=minimizer_supplier,
            minimizer=minimizer, dt=dt, eps=eps, mu=mu, N_steps=N_steps, N_epochs=N_epochs, 
            custom_keyboard_interrupt_handler=interrupt_handler,
            display_format=display_format, callbacks=callbacks
    )

    if verbose:
        print(f'elongating interface')


    left = fu.crop(model.field, 0, 0.5, 0, 1)
    right = fu.crop(model.field, 0.5, 1, 0, 1)

    tmp = fu.concat(delta_liq, left)
    tmp = fu.concat(tmp, delta_sol)    
    tmp = fu.concat(tmp, delta_sol)    
    tmp = fu.concat(tmp, right)   
    ifc_elongated = fu.concat(tmp, delta_liq)

    if verbose:
        print(f'Lx={ifc_elongated.Lx}, Ly={ifc_elongated.Ly}')
    return model, ifc_elongated


     



import numpy as np
from matplotlib import pyplot as plt
from scipy.fft import fft2, ifft2, fft, ifft
from pprint import pprint


def modify_and_save(new_file_name, saved, **modified):
    data = dict(saved)
    for m in modified:
        data[m] = modified[m]
    np.savez(new_file_name, **data)

def change_resolution(psi, Nx1, Ny1):
    Nx, Ny = psi.shape[0], psi.shape[1]
    psi1 = np.zeros((Nx1, Ny1))

    for i in range(Nx1):
        for j in range(Ny1):
            psi1[i,j] = psi[int(i*Nx/Nx1),int(j*Ny/Ny1)]
    return psi1

def periodic_extend(psi, Mx, My):
    Nx, Ny = psi.shape[0], psi.shape[1]
    Nx1 = Nx * Mx
    Ny1 = Ny * My

    psi1 = np.zeros((Nx1, Ny1))

    for i in range(Nx1):
        for j in range(Ny1):
            psi1[i,j] = psi[i%Nx,j%Ny]
    return psi1







import time
import threading
import sys
import warnings
from pprint import pprint
import tqdm
from typing import List, Optional

import numpy as np
from matplotlib import pyplot as plt
from scipy.fft import fft2, ifft2, rfft2, irfft2, set_global_backend
import pyfftw

from michael960lib.math import fourier
from michael960lib.common import overrides, experimental, deprecated
from michael960lib.common import ModifyingReadOnlyObjectError, IllegalActionError
from torusgrid.fields import RealField2D, import_field
from torusgrid.dynamics import FancyEvolver, NoiseGenerator2D, FreeEnergyFunctional2D, EvolverHistory
from .base import PFCStateFunction, PFCFreeEnergyFunctional, import_state_function




class PFCMinimizer(FancyEvolver):
    def __init__(self, field: RealField2D, dt, eps):
        super().__init__(field)
        self.dt = dt
        self.eps = eps
        self.age = 0
        self.history = PFCMinimizerHistory()

        self.info['system'] = 'pfc'
        self.info['minimizer'] = 'NULL'

        self.info['dt'] = dt
        self.info['eps'] = eps
        self.info['label'] = self.label = f'NULL eps={eps} dt={dt}'

        self.display_format = '[{label}] f={f:.5f} F={F:.5f} psibar={psibar:.5f}'
    
    @overrides(FancyEvolver)
    def get_evolver_state(self):
        return {'age': self.age}

    @overrides(FancyEvolver)
    def on_create_progress_bar(self, progress_bar: tqdm.tqdm):
        progress_bar.set_description_str(f'[{self.label}]')
        
    def set_age(self, age):
        self.age = age


class ConstantChemicalPotentialMinimizer(PFCMinimizer):
    def __init__(self, field: RealField2D, dt: float, eps: float, mu: float, noise_generator:Optional[NoiseGenerator2D]=None):
        super().__init__(field, dt, eps)
        self.info['minimizer'] = 'mu'
        self.info['mu'] = self.mu = mu
        self.info['label'] = self.label = f'mu eps={eps:.5f} mu={mu:.5f} dt={dt:.5f}'
   
        self.fef = PFCFreeEnergyFunctional(eps)
        self.noise_generator = noise_generator

        self.is_noisy = not (noise_generator is None)

        if not field.fft_initialized():
            field.initialize_fft() 
        self._kernel = 1-2*self.field.K2+self.field.K4 - self.eps
        self._exp_dt_kernel = np.exp(-dt*self._kernel)
        self._mu_dt_half = self.dt * self.mu / 2

    @overrides(PFCMinimizer)
    def step(self):
        self.age += self.dt
        if self.is_noisy:
            self.field.psi += self.dt * self.noise_generator.generate() 

        self.field.psi[:] += self._mu_dt_half
        self.field.psi /=np.sqrt(1+self.field.psi**2*self.dt)
        
        self.field.fft()
        self.field.psi_k *= self._exp_dt_kernel
        self.field.ifft()

        self.field.psi /= np.sqrt(1+self.field.psi**2*self.dt)
        self.field.psi[:] += self._mu_dt_half

    @overrides(PFCMinimizer)
    def get_state_function(self):
        psibar = np.mean(self.field.psi)
        psiN = psibar * self.field.Volume
        f = self.fef.mean_free_energy_density(self.field)
        F = self.fef.free_energy(self.field)
        omega = f - self.mu * psibar
        Omega = F - self.mu * psiN

        return PFCStateFunction(self.field.Lx, self.field.Ly, f, F, psibar, omega, Omega)


class NonlocalConservedMinimizer(PFCMinimizer):
    def __init__(self, field: RealField2D, dt: float, eps: float, noise_generator:NoiseGenerator2D=None):
        super().__init__(field, dt, eps)

        self.info['label'] = self.label = f'nonlocal eps={eps:.5f} dt={dt:.5f}'
        self.info['minimizer'] = 'nonlocal'

        self._kernel = 1-2*self.field.K2+self.field.K4 - self.eps
        self._exp_dt_kernel = np.exp(-dt*self._kernel)
        self.noise_generator = noise_generator
        self.fef = PFCFreeEnergyFunctional(eps)
        self.is_noisy = not (noise_generator is None)

        if not field.fft_initialized():
            field.initialize_fft()

        self.psibar = np.mean(self.field.psi)

    @overrides(PFCMinimizer)
    def step(self):
        self.age += self.dt

        if self.is_noisy:
            self.field.psi += self.dt * self.noise_generator.generate() 
        
        self.field.psi /= np.sqrt(1+self.field.psi**2*self.dt)
        self.field.psi += - np.mean(self.field.psi) + self.psibar

        self.field.fft()
        self.field.psi_k *= self._exp_dt_kernel
        self.field.ifft()
        self.field.psi += - np.mean(self.field.psi) + self.psibar

        self.field.psi /= np.sqrt(1+self.field.psi**2*self.dt)
        self.field.psi += - np.mean(self.field.psi) + self.psibar

    @overrides(PFCMinimizer)
    def get_state_function(self):
        psibar = np.mean(self.field.psi)
        f = self.fef.mean_free_energy_density(self.field)
        F = self.fef.free_energy(self.field)
        return PFCStateFunction(self.field.Lx, self.field.Ly, f, F, psibar)


class NonlocalConservedRK4(NonlocalConservedMinimizer):
    @experimental('Nonlocal RK4 is experimental and not optimized')
    def __init__(self, field: RealField2D, dt: float, eps: float,
            noise_generator: Optional[NoiseGenerator2D]=None,
            k_regularizer=0.1, inertia=100):
        super().__init__(field, dt, eps, noise_generator)
        self.R = k_regularizer
        self.info['minimizer'] = 'nonlocal-rk4'
        self.info['M'] = inertia
        self.info['R'] = k_regularizer
        self.info['label'] = self.label = f'nonlocal-rk4 eps={eps} dt={dt} R={k_regularizer} M={inertia}'

        self.field_tmp = self.field.copy()
        self.field_tmp.initialize_fft()

        self._deriv = RealField2D(field.Lx, field.Ly, field.Nx, field.Ny)
        self._deriv.initialize_fft()

        self.dfield = self.field.copy()
        self.dfield.set_psi(0)
        self.dfield.initialize_fft()

        self.dfield_tmp = self.dfield.copy()
        self.dfield_tmp.initialize_fft()

        self.inertia = inertia

    def _psi_dot(self):
        self._deriv.psi[:,:] = -self.fef.derivative(self.field_tmp)
        self._deriv.fft()
        self._deriv.psi_k *= np.exp(-self.R*self._deriv.K2)
        self._deriv.ifft()
        F = self._deriv.psi - np.mean(self._deriv.psi)
        return self.dfield_tmp.psi*self.inertia, -(self.dfield_tmp.psi - F)

    # RK4
    @overrides(NonlocalConservedMinimizer)
    def step(self):
        self.age += self.dt
        self.field_tmp.psi[:] = self.field.psi
        self.dfield_tmp.psi[:] = self.dfield.psi

        k1, l1 = self._psi_dot()
        self.field_tmp.psi += k1*self.dt/2
        self.dfield_tmp.psi += l1*self.dt/2

        k2, l2 = self._psi_dot()
        self.field_tmp.psi += k2*self.dt/2
        self.dfield_tmp.psi += l2*self.dt/2

        k3, l3 = self._psi_dot()
        self.field_tmp.psi += k3*self.dt
        self.dfield_tmp.psi += l3*self.dt

        k4, l4 = self._psi_dot()

        self.field.psi += self.dt / 6 * (k1 + 2*k2 + 2*k3 + k4)
        self.dfield.psi += self.dt / 6 * (l1 + 2*l2 + 2*l3 + l4)
 

class NonlocalConservedRK4Plain(NonlocalConservedRK4):
    def __init__(self, field, dt, eps,
            noise_generator: Optional[NoiseGenerator2D]=None,
            k_regularizer=0.1):
        super().__init__(field, dt, eps, noise_generator, k_regularizer, 0)

    def _psi_dot(self):
        self._deriv.psi[:,:] = -self.fef.derivative(self.field_tmp)
        self._deriv.fft()
        self._deriv.psi_k *= np.exp(-self.R*self._deriv.K2)
        self._deriv.ifft()
        F = self._deriv.psi - np.mean(self._deriv.psi)
        return F

    @overrides(NonlocalConservedMinimizer)
    def step(self):
        self.age += self.dt
        self.field_tmp.psi[:] = self.field.psi

        k1 = self._psi_dot()
        self.field_tmp.psi += k1*self.dt/2

        k2 = self._psi_dot()
        self.field_tmp.psi += k2*self.dt/2

        k3 = self._psi_dot()
        self.field_tmp.psi += k3*self.dt

        k4 = self._psi_dot()

        self.field.psi += self.dt / 6 * (k1 + 2*k2 + 2*k3 + k4)


class NonlocalDescent(NonlocalConservedMinimizer):
    def __init__(self, field: RealField2D, dt: float, eps: float,
            noise_generator: Optional[NoiseGenerator2D]=None,
            k_regularizer=0.1):
        super().__init__(field, dt, eps, noise_generator)
        self.R = k_regularizer
        self.info['minimizer'] = 'nonlocal-desc'
        self.info['R'] = k_regularizer
        self.info['label'] = self.label = f'nonlocal-desc eps={eps} dt={dt} R={k_regularizer}'

        self.field_tmp = self.field.copy()
        self.field_tmp.initialize_fft()

        self._deriv = RealField2D(field.Lx, field.Ly, field.Nx, field.Ny)
        self._deriv.initialize_fft()

    def _psi_dot(self):
        self._deriv.psi[:,:] = -self.fef.derivative(self.field_tmp)
        self._deriv.fft()
        self._deriv.psi_k *= np.exp(-self.R*self._deriv.K2)
        self._deriv.ifft()
        F = self._deriv.psi - np.mean(self._deriv.psi)
        return F

    @overrides(NonlocalConservedMinimizer)
    def step(self):
        self.age += self.dt
        self.field_tmp.psi[:,:] = self.field.psi[:,:]
        F = self._psi_dot()
        cutoff = 0.01 * self.dt

        dT = self.dt / (np.sum(F**2)*self.field.dV + cutoff)
        self.field.psi += F * dT


class NonlocalTeleporter(NonlocalConservedRK4):
    def __init__(self, dt, eps):
        super().__init__(field, dt, eps, inertia=0)

    @overrides(NonlocalConservedMinimizer)
    def step(self):
        raise NotImplementedError
 


class ConservedMinimizer(PFCMinimizer):
    @experimental('Local conserved minimizer is not implemented yet')
    def __init__(self, field: RealField2D, dt: float, eps: float, noise_generator:NoiseGenerator2D=None):
        super().__init__(field, dt, eps)
        self.label = f'nonlocal eps={eps} dt={dt}'

        self._kernel = 1-2*self.field.K2+self.field.K4 - self.eps
        self._exp_dt_kernel = np.exp(-dt*self._kernel)
        self.noise_generator = noise_generator
        self.fef = PFCFreeEnergyFunctional(eps)
        self.is_noisy = not (noise_generator is None)

        if not field.fft_initialized():
            field.initialize_fft()

        self.psibar = np.mean(self.field.psi)

        self.field_tmp = self.field.copy()

    #@overrides(PFCMinimizer)
    def step(self):
        raise NotImplementedError
        self.age += self.dt

        if self.is_noisy:
            self.field.psi += self.dt * self.noise_generator.generate() 
        
        self.field.psi /= np.sqrt(1+self.field.psi**2*self.dt)
        self.field.psi += - np.mean(self.field.psi) + self.psibar

        self.field.fft()
        self.field.psi_k *= self._exp_dt_kernel
        self.field.ifft()
        self.field.psi += - np.mean(self.field.psi) + self.psibar

        self.field.psi /= np.sqrt(1+self.field.psi**2*self.dt)
        self.field.psi += - np.mean(self.field.psi) + self.psibar



    @overrides(PFCMinimizer)
    def get_state_function(self):
        psibar = np.mean(self.field.psi)
        f = self.fef.mean_free_energy_density(self.field)
        F = self.fef.free_energy(self.field)
        return PFCStateFunction(self.field.Lx, self.field.Ly, f, F, psibar)


class StressRelaxer(PFCMinimizer):
    def __init__(self, field: RealField2D, dt: float, eps: float, mu: float,
            noise_generator: NoiseGenerator2D=None, expansion_rate: float=1):
        super().__init__(field, dt, eps)
        self.f = 1

        self.info['mu'] = mu
        self.info['label'] = self.label = f'stress-relax eps={eps:.5f} mu={mu:.5f} dt={dt:.5f}'
        self.info['minimizer'] = 'stress-relax'
        self.info['expansion_rate'] = expansion_rate

        self.dt = dt
        self.eps = eps
        self.mu = mu
        self.noise_generator = noise_generator
        self.is_noisy = not (noise_generator is None)
        self.fef = PFCFreeEnergyFunctional(eps)
        self.dT = self.dt * expansion_rate / 2
        self.NN = field.Nx * field.Ny

        if not field.fft_initialized():
            field.initialize_fft()

        self._prepare_minimization()


    def _prepare_minimization(self):
        f = self.field
        #self._exp_dt_eps_half = np.exp(self.dt*self.eps/2)
        self._mu_dt_half = self.dt * self.mu / 2
        #self._conv_helper = self.field.K2 * 0 + 2/self.NN
        self._2_NN = 2 / self.NN

        self._domega_kernels = np.array([
            2/f.Lx*f.Kx2*(1-f.K2), 
            2/f.Ly*f.Ky2*(1-f.K2), 
            -2/f.Lx**2*f.Kx2*(3-5*f.Kx2-3*f.Ky2),
            -2/f.Ly**2*f.Ky2*(3-5*f.Ky2-3*f.Kx2),
            2/f.Lx/f.Ly * 2*f.Kx2*f.Ky2
        ])

        self.Lx0 = f.Lx
        self.Ly0 = f.Ly

        self.set_size_scale(1., 1.)

    @overrides(PFCMinimizer)
    def step(self):
        f = self.field
        self.age += self.dt
        if self.is_noisy:
            f.psi += self.dt * self.noise_generator.generate() 

        f.psi += self._mu_dt_half
        f.psi /=np.sqrt(1+f.psi**2*self.dt)
        
        f.fft()

        _kernel = 1-2*self.K2+self.K4 - self.eps
        _exp_dt_kernel = np.exp(-self.dt*_kernel/2)
        f.psi_k *= _exp_dt_kernel

        self.relax_stress_full()

        _kernel = 1-2*self.K2+self.K4 - self.eps
        _exp_dt_kernel = np.exp(-self.dt*_kernel/2)
        f.psi_k *= _exp_dt_kernel

        f.ifft()

        f.psi /= np.sqrt(1+self.field.psi**2*self.dt)
        f.psi += self._mu_dt_half

    def relax_stress_full(self):
        f = self.field
        #f = self
        dT = self.dt

        self._domega_kernels[:,:,:] = [
                2/self.Lx*self.Kx2*(1-f.K2), 
                2/self.Ly*self.Ky2*(1-f.K2), 
                -2/self.Lx**2*self.Kx2*(3-5*self.Kx2-3*self.Ky2),
                -2/self.Ly**2*self.Ky2*(3-5*self.Ky2-3*self.Kx2),
                2/self.Lx/self.Ly * 2*self.Kx2*self.Ky2]

        omega_list = self.real_convolution_2d(np.abs(f.psi_k**2), self._domega_kernels)

        dLx = -omega_list[0]*dT + (omega_list[0] * omega_list[2] + omega_list[1] * omega_list[4]) * dT**2/2
        dLy = -omega_list[1]*dT + (omega_list[1] * omega_list[3] + omega_list[0] * omega_list[4]) * dT**2/2

        dfx = dLx / self.Lx0
        dfy = dLy / self.Ly0

        #f.set_size(f.Lx+dLx, f.Ly+dLy)
        self.set_size_scale(self.fx + dfx, self.fy + dfy)

    # size scale factor compared to the original dimensions
    def set_size_scale(self, fx, fy):
        self.fx = fx
        self.fy = fy
        self.Kx2 = self.field.Kx2 / fx**2
        self.Ky2 = self.field.Ky2 / fy**2
        self.K2 = self.Kx2 + self.Ky2
        self.K4 = self.K2**2
        self.Lx = self.field.Lx * self.fx
        self.Ly = self.field.Ly * self.fy


    @overrides(PFCMinimizer)
    def get_state_function(self):
        self.field.set_size(self.Lx, self.Ly)
        self.set_size_scale(1., 1.)
        psibar = np.mean(self.field.psi)
        psiN = psibar * self.field.Volume

        f = self.fef.mean_free_energy_density(self.field)
        F = self.fef.free_energy(self.field)

        omega = f - self.mu * psibar
        Omega = F - self.mu * psiN

        return PFCStateFunction(self.field.Lx, self.field.Ly, f, F, psibar, omega, Omega)

    def real_convolution_2d(self, psi_k_sq, kernel):
        r = kernel * psi_k_sq
        r[:,:,[0,-1]] *= 0.5
        return r.sum(axis=(1,2)) * self._2_NN


class RandomStepMinimizer(PFCMinimizer):
    def __init__(self, field: RealField2D, dt: float, eps: float, mu: float, seed: int):
        super().__init__(field, dt, eps)
        self.target = 'Omega'

        self.info['minimizer'] = 'random'
        self.info['label'] = self.label = 'random eps={eps:.5f} mu={mu:.5f} dt={dt:.5f}'

        if mu is None:
            self.target = 'F'
            self.label = 'random eps={eps} dt={dt}'

        self.mu = mu
        self.fef = PFCFreeEnergyFunctional(eps)

        self.generator = np.random.default_rng(seed)
        self.candidate_field = self.field.copy()

        self.random_scale = self.dt**2 * np.sqrt(np.mean(self.fef.derivative(self.field)**2))
        print(f'scale = {self.random_scale}')

    @overrides(PFCMinimizer) 
    def step(self):
        self.age += self.dt
        self.candidate_field.psi = self.field.psi + self.generator.normal(0, self.random_scale, size=(self.field.Nx,self.field.Ny))
        if self.target == 'F':
            F0 = self.fef.free_energy(self.field) 
            F1 = self.fef.free_energy(self.candidate_field)
            if F1 < F0:
                self.field.psi[:,:] = self.candidate_field.psi
        if self.target == 'Omega':
            O0 = self.fef.grand_potential(self.field, self.mu) 
            O1 = self.fef.grand_potential(self.candidate_field, self.mu)
            if O1 < O0:
                self.field.psi[:,:] = self.candidate_field.psi


    @overrides(PFCMinimizer)
    def get_state_function(self):
        Lx = self.field.Lx
        Ly = self.field.Ly
        psibar = np.mean(self.field.psi)
        f = self.fef.mean_free_energy_density(self.field)
        F = self.fef.free_energy(self.field)

        omega = None
        Omega = None
        if self.target == 'Omega':
            omega = self.fef.mean_grand_potential_density(self.field, self.mu)
            Omega = self.fef.grand_potential(self.field, self.mu)

        return PFCStateFunction(Lx, Ly, f, F, psibar, omega, Omega)




class PFCMinimizerHistory(EvolverHistory):
    def __init__(self):
        super().__init__()
        self.t = []
        self.age = 0

    @overrides(EvolverHistory)
    def append_state_function(self, evolver_state: dict, sf: PFCStateFunction):
        if self.committed:
            raise ModifyingReadOnlyObjectError(
            f'history object (label=\'{self.label}\') is already committed and hence not editable', self)

        t = evolver_state['age']
        if t < self.age:
            raise IllegalActionError(f'time={t} is smaller than the current recorded time')

        self.t.append(t)
        self.state_functions.append(sf)
        self.age = t

    def get_t(self):
        return self.t

    def export(self) -> dict:
        if self.committed:
            state = dict()
            state['age'] = self.age
            state['label'] = self.label
            state['final_field_state'] = self.final_field_state
            state['state_functions'] = [sf.export() for sf in self.state_functions]
            state['t'] = self.t
            return state
        else:
            raise IllegalActionError(
            'history object (label=\'{self.label}\') has not been committed and is therefore not ready to be exported')


def import_minimizer_history(state: dict) -> PFCMinimizerHistory:
    mh = PFCMinimizerHistory()
    mh.label = state['label']
    mh.state_functions = [import_state_function(sf_state) for sf_state in state['state_functions']]
    mh.age = state['age']
    mh.final_field_state = state['final_field_state']
    mh.t = state['t']
    mh.committed = True
    return mh






from typing import Optional

import numpy as np
from scipy.fft import rfft2, irfft2

from michael960lib.common import overrides, deprecated, experimental
from torusgrid.fields import RealField2D
from torusgrid.dynamics import FreeEnergyFunctional2D, NoiseGenerator2D
from ..core.base import PFCFreeEnergyFunctional
from ..core.evolution import ConstantChemicalPotentialMinimizer, StressRelaxer, PFCMinimizer, NonlocalConservedRK4


class PFCFunctional6(PFCFreeEnergyFunctional):
    def __init__(self, eps: float, alpha: float):
        self.eps = eps
        self.alpha = alpha

    @overrides(FreeEnergyFunctional2D)
    def free_energy_density(self, field: RealField2D):
        kernel = (1-field.K2)**2 +\
                self.alpha*(field.K4*field.K2 - 2*field.K4 + field.K2)
        psi_k = rfft2(field.psi)
        psi_k_o = kernel * psi_k
        f = 1/2 * field.psi * irfft2(psi_k_o) + field.psi**4/4 - self.eps/2 * field.psi**2
        return np.real(f)


    @overrides(FreeEnergyFunctional2D)
    def derivative(self, field):
        field.fft()
        linear_term = ((1-field.K2)**2 + self.alpha*field.K2*(1-field.K2)**2 - self.eps) * field.psi_k
        field.ifft()

        local_mu = irfft2(linear_term) + field.psi**3
        return local_mu



class PFC6Minimizer(ConstantChemicalPotentialMinimizer):
    def __init__(self,
            field: RealField2D, dt: float, 
            eps: float, alpha: float, mu: float, 
            noise_generator:Optional[NoiseGenerator2D]=None):

        super().__init__(field, dt, eps, mu, noise_generator)
        self.alpha = alpha
        self.info['system'] = 'pfc6'
        self.info['alpha'] = alpha
        self.info['label'] = self.label = f'mu6 eps={eps:.5f} mu={mu:.5f} alpha={alpha:.5f} dt={dt:.5f}'
        
        self.fef = PFCFunctional6(eps, alpha)

        _kernel = 1 - 2*field.K2 + field.K4 +\
                alpha*(field.K4*field.K2 -2*field.K4 + field.K2)- eps

        self._exp_dt_kernel = np.exp(-dt*_kernel)
        self._mu_dt_half = dt * mu / 2



class PFC6Relaxer(StressRelaxer):
    def __init__(self, field: RealField2D, dt: float, eps: float, alpha: float, mu: float, expansion_rate: float=1.):
        super().__init__(field, dt, eps, mu, expansion_rate=expansion_rate)
        self.info['system'] = 'pfc6'
        self.info['alpha'] = alpha
        self.info['label'] = self.label = f'stress-relax6 eps={eps:.5f} mu={mu:.5f} dt={dt:.5f}'

        self.alpha = alpha
        self.fef = PFCFunctional6(eps, alpha)

    def _prepare_minimization(self):
        f = self.field
        self._mu_dt_half = self.dt * self.mu / 2
        self._2_NN = 2 / self.NN

        self._domega_kernels = np.array([
            2/f.Lx*f.Kx2*(1-f.K2), 
            2/f.Ly*f.Ky2*(1-f.K2), 
            -2/f.Lx**2*f.Kx2*(3-5*f.Kx2-3*f.Ky2),
            -2/f.Ly**2*f.Ky2*(3-5*f.Ky2-3*f.Kx2),
            2/f.Lx/f.Ly * 2*f.Kx2*f.Ky2
        ])

        self.Lx0 = f.Lx
        self.Ly0 = f.Ly

        self.Lx = f.Lx
        self.Ly = f.Ly
        self.set_size_scale(1., 1.)

    @overrides(PFCMinimizer)
    def step(self):
        f = self.field
        self.age += self.dt
        if self.is_noisy:
            f.psi += self.dt * self.noise_generator.generate() 

        f.psi += self._mu_dt_half
        f.psi /=np.sqrt(1+f.psi**2*self.dt)
        
        f.fft()

        _kernel = 1-2*self.K2+self.K4 - self.eps +\
                self.alpha*(self.K6 -2*self.K4 + self.K2)

        _exp_dt_kernel = np.exp(-self.dt*_kernel/2)
        f.psi_k *= _exp_dt_kernel

        self.relax_stress_full()

        _kernel = 1-2*self.K2+self.K4 - self.eps +\
                self.alpha*(self.K6 -2*self.K4 + self.K2)

        _exp_dt_kernel = np.exp(-self.dt*_kernel/2)
        f.psi_k *= _exp_dt_kernel


        f.ifft()

        f.psi /= np.sqrt(1+self.field.psi**2*self.dt)
        f.psi += self._mu_dt_half

    def relax_stress_full(self):
        f = self.field
        dT = self.dt

        self._domega_kernels[:,:,:] = [
                # domega/dLx
                1/self.Lx * self.Kx2 * (1-self.K2) * (2-self.alpha*(1-3*self.K2)),
                # domega/dLy
                1/self.Ly * self.Ky2 * (1-self.K2) * (2-self.alpha*(1-3*self.K2)),
                # d2omega/dLx2
                -1/self.Lx**2*self.Kx2 *\
                (6-10*self.Kx2-6*self.Ky2 - self.alpha*(3 - 20*self.Kx2 - 12*self.Ky2 + 21*self.Kx4 + 30*self.Kx2*self.Ky2 + 9*self.Ky4)),
                # d2omega/dLy2
                -1/self.Ly**2*self.Ky2 *\
                (6-10*self.Ky2-6*self.Kx2 - self.alpha*(3 - 20*self.Ky2 - 12*self.Kx2 + 21*self.Ky4 + 30*self.Ky2*self.Kx2 + 9*self.Kx4)),
                # d2omega/dLxdLy
                1/self.Lx/self.Ly * 4*self.Kx2*self.Ky2 * (1-self.alpha*(2-3*self.K2))
        ]

        omega_list = self.real_convolution_2d(np.abs(f.psi_k**2), self._domega_kernels)

        dLx = -omega_list[0]*dT + (omega_list[0] * omega_list[2] + omega_list[1] * omega_list[4]) * dT**2/2
        dLy = -omega_list[1]*dT + (omega_list[1] * omega_list[3] + omega_list[0] * omega_list[4]) * dT**2/2

        dfx = dLx / self.Lx0
        dfy = dLy / self.Ly0

        #f.set_size(f.Lx+dLx, f.Ly+dLy)
        self.set_size_scale(self.fx + dfx, self.fy + dfy)

    # size scale factor compared to the original dimensions
    def set_size_scale(self, fx, fy):
        self.fx = fx
        self.fy = fy

        self.Kx2 = self.field.Kx2 / fx**2
        self.Ky2 = self.field.Ky2 / fy**2
        self.K2 = self.Kx2 + self.Ky2

        self.Kx4 = self.Kx2**2
        self.Ky4 = self.Ky2**2
        self.K4 = self.K2**2

        self.K6 = self.K2**3

        self.Lx = self.field.Lx * self.fx
        self.Ly = self.field.Ly * self.fy


class PFC6NonlocalConservedRK4(NonlocalConservedRK4):
    def __init__(self, field: RealField2D, dt: float, eps: float, alpha: float,
            noise_generator: Optional[NoiseGenerator2D]=None,
            k_regularizer=0.1, inertia=100):
        super().__init__(field, dt, eps, noise_generator, k_regularizer, inertia)
        self.alpha = alpha
        self.info['alpha'] = alpha
        self.info['label'] = self.label = f'pfc6 nonlocal-rk4 eps={eps} alpha={alpha} dt={dt} R={k_regularizer} M={inertia}'
        self.fef = PFCFunctional6(eps, alpha)

__all__ = ['PFCFunctional6', 'PFC6Minimizer']








import numpy as np
from matplotlib import pyplot as plt
from scipy.fft import rfft2, irfft2
import pyfftw

from pprint import pprint
import tqdm
import time
import threading
import sys
from typing import List, Optional

from michael960lib.math import fourier
from michael960lib.common import overrides, ModifyingReadOnlyObjectError, IllegalActionError
from torusgrid.fields import RealField2D, import_field, FieldStateFunction
from torusgrid.dynamics import FreeEnergyFunctional2D, NoiseGenerator2D


class PFCFreeEnergyFunctional(FreeEnergyFunctional2D):
    def __init__(self, eps):
        self.eps = eps

    @overrides(FreeEnergyFunctional2D)
    def free_energy_density(self, field: RealField2D):
        kernel = 1-2*field.K2+field.K4
        psi_k = rfft2(field.psi)
        psi_k_o = kernel * psi_k
        f = 1/2 * field.psi * irfft2(psi_k_o) + field.psi**4/4 - self.eps/2 * field.psi**2
        return np.real(f)

    @overrides(FreeEnergyFunctional2D)
    def derivative(self, field: RealField2D):
        field.fft()
        linear_term = ((1-field.K2)**2 - self.eps) * field.psi_k
        field.ifft()

        #local_mu = (1-self.eps) * field.psi + field.psi**3 + 2*D2psi + D4psi
        local_mu = irfft2(linear_term) + field.psi**3
        return local_mu

    def grand_potential_density(self, field: RealField2D, mu: float):
        return self.free_energy_density(field) - mu*field.psi

    def mean_grand_potential_density(self, field: RealField2D, mu: float):
        return np.mean(self.grand_potential_density(field, mu))

    def grand_potential(self, field: RealField2D, mu: float):
        return np.sum(self.grand_potential_density(field, mu)) * field.dV

class PFCStateFunction(FieldStateFunction):
    def __init__(self, Lx, Ly, f, F, psibar, omega=None, Omega=None):
        super().__init__()
        self.Lx = self._content['Lx'] = Lx
        self.Ly = self._content['Ly'] = Ly
        self.f = self._content['f'] = f
        self.F = self._content['F'] = F
        self.psibar = self._content['psibar'] = psibar
        self.omega = self._content['omega'] = omega
        self.Omega = self._content['Omega'] = Omega

    def is_grand_canonical(self) -> bool:
        return not (self.omega is None)

    def to_string(self, state_string_format: Optional[str]=None,
            float_fmt: str='.3f', pad: int=1, delim: str='|') -> str:
        if not state_string_format is None:
            return state_string_format.format(
                Lx=self.Lx, Ly=self.Ly, f=self.f, F=self.F, 
                omega=self.omega, Omega=self.Omega, psibar=self.psibar
            )

        delim_padded = ' '*pad + delim + ' '*pad
        items = ['Lx', 'Ly', 'f', 'F', 'omega', 'Omega', 'psibar']
        state_func_list = []

        for item_name in items:
            item = self.item_dict[item_name]
            if not item is None: 
                state_func_list.append(f'{item_name}={item:{float_fmt}}')
        return delim_padded.join(state_func_list)


def import_state_function(state: dict) -> PFCStateFunction:
    sf = PFCStateFunction(state['Lx'], state['Ly'], state['f'], state['F'], state['psibar'], state['omega'], state['Omega'])
    return sf

def get_latex(item_name) -> str:
    if not item_name in _item_latex_dict.keys():
        raise ValueError(f'{item_name} is not a valid state function function')
    return _item_latex_dict[item_name]

_item_latex_dict = {
    'Lx': r'$Lx$',
    'Ly': r'$Ly$',
    'f': r'$f$',
    'F': r'$F$',
    'omega': r'$\omega$',
    'Omega': r'$\Omega$',
    'psibar': r'$\bar\psi$'
}





import numpy as np
from matplotlib import pyplot as plt
from scipy.fft import fft2, ifft2
from pprint import pprint
import pyfftw
import time
import threading
import sys

from michael960lib.math import fourier

import warnings

warnings.warn('the pfc module is deprecated, use pfc_im instead')

class PhaseFieldCrystal2D:
    def __init__(self, Lx, Ly, Nx, Ny):

        self.set_dimensions(Lx, Ly, Nx, Ny, reset_psi=True)
        self.set_psi(0)
        self.set_params(0, 0)
        self.set_noise(0, 0)

        self.lock = None


    def set_noise(self, seed, width):

        self.seed = seed
        np.random.seed(self.seed)

        self.noise_width = width

    def set_params(self, mu, eps):
        self.mu = mu
        self.eps = eps
        print(f'[pfc] new params set, mu={mu} eps={eps}')

    def set_dimensions(self, Lx, Ly, Nx, Ny, reset_psi=True):

        self.Lx = Lx
        self.Ly = Ly
        self.Volume = Lx*Ly

        self.Nx = Nx
        self.Ny = Ny
        
        x, kx, self.dx, self.dkx = fourier.generate_xk(Lx, Nx)
        y, ky, self.dy, self.dky = fourier.generate_xk(Ly, Ny)
        
        self.X, self.Y = np.meshgrid(x, y, indexing='ij')
        self.Kx, self.Ky = np.meshgrid(kx, ky, indexing='ij')
        self.K2 = self.Kx**2 + self.Ky**2
        self.K4 = self.K2**2

        self.Kernel = 1-2*self.K2+self.K4
        
        print(f'[pfc] new dimensions set, Lx={Lx} Ly={Ly} Nx={Nx} Ny={Ny}')

        if reset_psi:
            self.psi = np.zeros((Nx, Ny))
            print(f'[pfc] resetted psi')



   
    def set_psi(self, psi1, verbose=True):
        self.psi[:,:] = psi1
        if verbose:
            print('[pfc] new psi set')

    def load_profile_from_file(self, saved_npz):
        file_list = saved_npz.files
        print(f'[pfc] loading npz file, found the following data:')
        print(f'[pfc] {file_list}')

        psi = saved_npz['psi']

        self.set_dimensions(saved_npz['Lx'], saved_npz['Ly'], *psi.shape)
        self.set_params(saved_npz['mu'], saved_npz['eps'])
        self.set_psi(saved_npz['psi'])
        

    def _evolve_k(self, dt):
        psik = fft2(self.psi)
        psik *= np.exp(-dt*self.Kernel)
        self.psi = np.real(ifft2(psik))

    def _evolve_x(self, dt):
        self.psi *= np.exp(-dt * (-self.eps + self.psi**2))
        self.psi += dt * self.mu

    def _evolve_conserved_nonlocal(self, dt):
        psi3k = fft2(self.psi**3)
        psik = fft2(self.psi)

        psik00 = psik[0,0]
        psik *= np.exp(-dt*(self.Kernel-self.eps))
        psik += -dt * psi3k
        psik[0,0] = psik00
        self.psi = np.real(ifft2(psik))


    def _evolve_noise(self, dt):
        self.psi += dt * np.random.normal(0, self.noise_width, size=(self.Nx, self.Ny))

    # total particl number = psi integrated over all volume
    def calc_N_tot(self):
        return np.sum(np.real(self.psi)) * self.dx * self.dy

    # local chemical potential
    def calc_chemical_potential_density(self):
        D2psi = ifft2(-self.K2*fft2(self.psi))
        D4psi = ifft2(self.K4*fft2(self.psi))
        local_mu = (1-self.eps) * self.psi + self.psi**3 + 2*D2psi + D4psi

        return np.real(local_mu)

    # local grand potential density
    def calc_grand_potential_density(self):
        return self.calc_helmholtz_density() - self.mu * self.psi

    def calc_grand_potential(self):
        return np.sum(self.calc_grand_potential_density()) * self.dx * self.dy
    
    # local helmholtz free energy density
    def calc_helmholtz_density(self):
        psi_k = fft2(self.psi)
        psi_k_o = self.Kernel * psi_k
        f = 1/2 * self.psi * ifft2(psi_k_o) + self.psi**4/4 - self.eps/2 * self.psi**2
        return np.real(f)

    #
    def calc_ord_param(self):
        return np.max(self.psi) - np.min(self.psi)

    
    # plot psi, chemical potential density, and grand potential density
    def plot(self, cmap='jet', plot_psi=True, plot_mu=True, plot_omega=True, lazy_factor=1):
        plt.figure(figsize=(12.8, 9.6))
        
        num_plots = plot_psi + plot_mu + plot_omega
        index_plot = 1


        LF = lazy_factor

        if plot_psi:
            ax1 = plt.subplot(num_plots, 1, index_plot, title='$\\psi$')
            cm1 = ax1.pcolormesh(self.X[::LF], self.Y[::LF], np.real(self.psi)[::LF], cmap=cmap, vmin=-1, vmax=1, shading='nearest')
            plt.colorbar(cm1, ax=ax1, orientation='horizontal')
            ax1.set_aspect('equal', adjustable='box')

            index_plot += 1

        if plot_mu:

            ax2 = plt.subplot(num_plots, 1, index_plot, title='$\\mu(\\mathbf{r})-\\mu$')
            cm2 = ax2.pcolormesh(self.X[::LF], self.Y[::LF], self.calc_chemical_potential_density()[::LF] - self.mu, cmap=cmap, vmin=-0.2, vmax=0.2, shading='nearest')
            plt.colorbar(cm2, ax=ax2, orientation='horizontal')
            ax2.set_aspect('equal', adjustable='box')

            index_plot += 1

        if plot_omega:

            ax3 = plt.subplot(num_plots, 1, index_plot, title='$\\omega(\\mathbf{r})$')
            cm3 = ax3.pcolormesh(self.X[::LF], self.Y[::LF], self.calc_grand_potential_density()[::LF], cmap=cmap, vmin=-0.1, vmax=0.1, shading='nearest')
            plt.colorbar(cm3, ax=ax3, orientation='horizontal')
            ax3.set_aspect('equal', adjustable='box')

            index_plot += 1


        plt.tight_layout()
        plt.show(block=True)


    def new_lock(self):
        self.lock = threading.Lock()
        return self.lock

    # minimize with constant mu
    def minimize_mu(self, dt, cycle=31, verbose=True, t0=0):
        i = 0
        t = t0
        if verbose:

            print(f'[pfc] ------------------------------------------------------------')
            print(f'[pfc] minimization parameter summary:')
            print(f'[pfc] mu={self.mu} eps={self.eps}')
            print(f'[pfc] Lx={self.Lx} Ly={self.Ly}')
            print(f'[pfc] Nx={self.Nx} Ny={self.Ny}')
            print(f'[pfc] noise_seed={self.seed} noise_amplitude={self.noise_width}')
            print(f'[pfc] dt={dt}')
            print(f'[pfc] ------------------------------------------------------------')
            print(f'[pfc] minimizing grand potential with constant mu={self.mu}')

        while True:
            with self.lock:
                self._evolve_noise(dt)
                self._evolve_x(dt/2)
                self._evolve_k(dt)
                self._evolve_x(dt/2)


            t += dt
            i += 1

            if i >= cycle:

                psi0 = self.calc_N_tot() / self.Volume
                omega = self.calc_grand_potential_density()
                Omega = np.sum(omega) * self.dx * self.dy
                ord_param = self.calc_ord_param()

                i = 0
                sys.stdout.write(f'\r[pfc] t={t:.4f} | psi_bar={psi0:.5f} | Omega={Omega:.5f} | diff={ord_param:.5f}    ')
        print()

    # minimize with nonlocal conserved dynamics
    def minimize_nonlocal_conserved(self, dt, cycle=31, verbose=True, t0=0):
        i = 0
        t = t0
        if verbose:
            psi0 = self.calc_N_tot() / self.Volume
            print(f'[pfc] minimizing grand potential with nonlocal conserved dynamics')
            print(f'[pfc] psi_bar={psi0}')

        while True:
            with self.lock:
                self._evolve_noise(dt)
                self._evolve_conserved_nonlocal(dt)


            t += dt
            i += 1

            if i >= cycle:

                psi0 = self.calc_N_tot() / self.Volume
                omega = self.calc_grand_potential_density()
                Omega = np.sum(omega) * self.dx * self.dy
                ord_param = self.calc_ord_param()

                i = 0
                sys.stdout.write(f'\r[pfc] t={t} | psi_bar={psi0} | Omega={Omega} | diff={ord_param}    ')
        print()




    def run_background(self, func, args):
        if self.lock is None:
            self.new_lock()
        
        thread = threading.Thread(target=func, args=args, daemon=True)
        thread.start()

        return self.lock, thread


class PhaseFieldCrystalExperimental(PhaseFieldCrystal2D):
    def __init__(self, Lx, Ly, Nx, Ny):
        super().__init__(Lx, Ly, Nx, Ny)
        
        self.set_energy_coeff(1, 0)

    def set_energy_coeff(self, p1, p2):
        self.p1 = p1
        self.p2 = p2

    def _evolve_x(self, dt):
        self.psi *= np.exp(-dt * (-self.eps + self.p1*self.psi**2 + self.p2*self.psi**4))
        self.psi += dt * self.mu


    def calc_chemical_potential_density(self):
        D2psi = ifft2(-self.K2*fft2(self.psi))
        D4psi = ifft2(self.K4*fft2(self.psi))
        local_mu = (1-self.eps) * self.psi + self.p1*self.psi**3 + self.p2*self.psi**5 + 2*D2psi + D4psi

        return np.real(local_mu)


    # local grand potential density
    def calc_grand_potential_density(self):
        return self.calc_helmholtz_density() - self.mu * self.psi
    
    # local helmholtz free energy density
    def calc_helmholtz_density(self):
        psi_k = fft2(self.psi)
        psi_k_o = self.Kernel * psi_k
        f = 1/2 * self.psi * ifft2(psi_k_o) + self.p1*self.psi**4/4 + self.p2*self.psi**6/6 - self.eps/2 * self.psi**2
        return np.real(f)

    # plot psi, chemical potential density, and grand potential density
    def plot(self, cmap='jet'):
        plt.figure(figsize=(12.8, 9.6))
        ax1 = plt.subplot(311, title='$\\psi$')
        ax2 = plt.subplot(312, title='$\\mu(\\mathbf{r})-\\mu$')
        ax3 = plt.subplot(313, title='$\\omega(\\mathbf{r})$')

        cm1 = ax1.pcolormesh(self.X, self.Y, np.real(self.psi), cmap=cmap, shading='nearest')
        ax1.set_aspect('equal', adjustable='box')

        print(np.min(self.psi))
        print(np.max(self.psi))
        cm2 = ax2.pcolormesh(self.X, self.Y, self.calc_chemical_potential_density() - self.mu, cmap=cmap, shading='nearest')
        ax2.set_aspect('equal', adjustable='box')

        cm3 = ax3.pcolormesh(self.X, self.Y, self.calc_grand_potential_density(), cmap=cmap, vmin=-0.1, vmax=0.1, shading='nearest')
        ax3.set_aspect('equal', adjustable='box')

        plt.colorbar(cm1, ax=ax1, orientation='horizontal')
        plt.colorbar(cm2, ax=ax2, orientation='horizontal')
        plt.colorbar(cm3, ax=ax3, orientation='horizontal')
        plt.tight_layout()
        plt.show(block=True)






from ..toolkit import profile_editor as pe
from ..toolkit import ortho_lattice_generator as olg
from ..legacy import pfc_im
from matplotlib import pyplot as plt
import numpy as np
import sys
import subprocess
from michael960lib import common as cmn
from michael960lib.common import overrides
import os
import re
import readline
import numexpr
from .prompt import Command, CommandExecutionError, SimplePrompt
import time

_version = '2.6'
_date = '2022/04/13'
_author = 'michael'



class ProfileEditorPrompt(SimplePrompt):
    def __init__(self):
        super().__init__()
        self.prefix = '[PIPE]>> '
        self.add_command(CommandHelp())

        self.add_command(CommandLs())
        self.add_command(CommandListModels())

        self.add_command(CommandLoad())
        self.add_command(CommandSave())

        self.add_command(CommandView())
        self.add_command(CommandInfo())

        self.add_command(CommandExtend())
        self.add_command(CommandCrop())
        self.add_command(CommandChangeRes())
        self.add_command(CommandCopy())
        self.add_command(CommandDelete())
        self.add_command(CommandMove())

        self.add_command(CommandSetParam())
        self.add_command(CommandSetSize())
        self.add_command(CommandSetMinimizer())

        self.add_command(CommandInterface())
        self.add_command(CommandLiquefy())
        self.add_command(CommandInsert())
        self.add_command(CommandFlip())
        self.add_command(CommandRotate())
        self.add_command(CommandConcat())

        self.add_command(CommandExit())

        self.add_command(CommandHistory())
        self.add_command(CommandIPython())
        self.add_command(CommandResetHistory())
        self.add_command(CommandEvolve())

        self.memory_models = dict()

        try:
            readline.read_history_file('.history')
        except FileNotFoundError as e:
            self.output('new command history file initialized in ./.history')

        readline.set_history_length(256)

    @overrides(SimplePrompt)
    def put_banner(self):
        self.output('PFC Interactive Profile Editor (PIPE)')
        self.output(f'version: {_version}')
        self.output(f'last updated on {_date} by {_author}')
        print()
    
    @overrides(SimplePrompt)
    def output(self, s, level=0):
        color = ''
        if level == 0:
            color = cmn.bcolors.OKCYAN
        elif level == 1:
            color = cmn.bcolors.WARNING
        elif level == 2:
            color = cmn.bcolors.FAIL
        print(f'{cmn.bcolors.BOLD}{color}[PIPE]{cmn.bcolors.ENDC} {s}')


    def get(self, varname) -> pfc_im.PhaseFieldCrystal2D:
        try:
            return self.memory_models[varname]
        except KeyError:
            s = f'model {varname} not found'
            raise CommandExecutionError(s)

    def put(self, varname, model):
        self.check_varname(varname)
        self.memory_models[varname] = model

    @overrides(SimplePrompt)
    def _on_exit(self):
        readline.write_history_file('.history')
        return


    def check_varname(self, varname: str):
        blacklist = [
            varname.find('-') != -1,
            varname.find('+') != -1,
            varname.find('*') != -1,
            varname.find('/') != -1,
            varname.find('.') != -1,
            varname.find('!') != -1,
            varname.find('@') != -1,
            varname.find('(') != -1,
            varname.find(')') != -1,
            varname.find('#') != -1,
            varname.find('^') != -1,
            varname.find('&') != -1,
            varname.find('$') != -1,
            varname.find('%') != -1,
            varname[0].isnumeric()
        ]
        if any(blacklist):
            raise CommandExecutionError(f'invalid variable name: {varname}')


class CommandHelp(Command):
    def __init__(self):
        super().__init__('help')

    @overrides(Command)
    def execute(self, prompt: SimplePrompt, args: list):
        prompt.output('PFC profile editor, please enter one of the following commands')
        prompt.output(prompt.commands_str)
        print()

    @overrides(Command)
    def complete(self, prompt: SimplePrompt, args: list):
        return [] 


class CommandLs(Command):
    def __init__(self):
        super().__init__('ls')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        #prompt.output('listing files')
        p = subprocess.run(['ls'] + args)
        print()
        
    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) <= 1:
            return super().complete(prompt, args)
        return [] 


class CommandListModels(Command):
    def __init__(self):
        super().__init__('list-models')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        s = ''
        for md in prompt.memory_models:
            s = s + md + ' '
        prompt.output(f'{s}')
        print()
        
    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        return []


class CommandLoad(Command):
    def __init__(self):
        super().__init__('load')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 2:
            raise CommandExecutionError('syntax: load FILENAME VARNAME')
        filename = args[0]
        varname = args[1]

        try:
            saved = np.load(filename, allow_pickle=True)
        except FileNotFoundError:
            raise CommandExecutionError(f'file \'{filename}\' not found')

        model = pfc_im.PhaseFieldCrystal2D(1, 1, 1, 1, verbose=False)
        model.load_profile_from_file(saved, verbose=False)

        prompt.put(varname, model)
        prompt.output(f'loaded file {filename} to model {varname}')
        print()

        
    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1:
            return super().complete(prompt, args)
        if len(args) == 2:
            return _match_models(prompt, args[-1])
        return [] 


class CommandSave(Command):
    def __init__(self):
        super().__init__('save')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 2:
                raise CommandExecutionError('syntax: save VARNAME FILENAME')

        varname = args[0]
        filename = args[1]
        
        model = prompt.get(varname)
        model.dump_profile_to_file(filename, verbose=False)

        prompt.output(f'saved model {varname} to file {filename}')
        print()
                
    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1:
            return _match_models(prompt, args[-1])
        if len(args) == 2:
            return super().complete(prompt, args)
        return [] 


class CommandView(Command):
    def __init__(self):
        super().__init__('view')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 1:
            raise CommandExecutionError('syntax: view VARNAME')

        varname = args[0]
        extra = args[1:]
        i = 0
        params = {'lazy': 1}
        try:
            while i < len(extra):
                if extra[i] == '--lazy':
                    assert i+1 < len(extra)
                    params['lazy'] = int(extra[i+1])
                    i += 1
                i += 1
        except AssertionError:
            prompt.output('error occured while parsing arguments:', level=1)
            prompt.output(f'{extra}', level=1)

        model = prompt.get(varname)

        prompt.output('invoking pfc model methods')
        model.plot(lazy_factor=params['lazy'], plot_psi=True, plot_mu=False, plot_omega=False)
        print()
                
    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1:
            return _match_models(prompt, args[-1])
        if len(args) == 2:
            return ['--lazy ']

        return [] 


class CommandInfo(Command):
    def __init__(self):
        super().__init__('info')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 1: 
            raise CommandExecutionError('syntax: info VARNAME [options]')
        varname = args[0]
        extra = args[1:]
        params = {'less': False, 'display-precision': 5}
        # parse extra args
        i = 0
        try:
            while i < len(extra):
                if extra[i] == '--less':
                    params['less'] = True
                elif extra[i] == '--display-precision':
                    assert i+1 < len(extra)
                    params['display-precision'] = int(extra[i+1])
                    i += 1
                i += 1

        except AssertionError:
            raise CommandExecutionError('error ocurred while parsing arguments')
        except ValueError as e:
            raise CommandExecutionError(e.args[-1])



        model = prompt.get(varname)
        dp = params['display-precision']
        mu = model.mu
        eps = model.eps
        Lx = model.Lx
        Ly = model.Ly
        Nx = model.Nx
        Ny = model.Ny
        Omega = model.calc_grand_potential()
        omega = Omega / model.Volume

        if params['less']:
            prompt.output(f'mu={mu}\teps={eps}')
            prompt.output(f'Lx={Lx:.{dp}f}\tLy={Ly:.{dp}f}\tNx={Nx}\tNy={Ny}')
            prompt.output(f'Omega={Omega:.{dp}f}\tomega={omega:.{dp}f}')
        else:
            model.summarize(display_precision=dp)
        print()
                
    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1:
            return _match_models(prompt, args[-1])
        if len(args) >= 1:
            if args[-1] == '--display-precision':
                return []
            return ['--display-precision ', '--less ']
        return [] 


class CommandExtend(Command):
    def __init__(self):
        super().__init__('extend')


    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 4:
            raise CommandExecutionError('syntax: extend VAR1 X-RATIO Y-RATIO VAR2')

        var1 = args[0]
        try:
            Mx = int(args[1])
            My = int(args[2])
        except ValueError as e:
            raise CommandExecutionError(e.args[-1])
        var2 = args[3]
        extra = args[4:]

        model1 = prompt.get(var1)
        model2 = model1.copy(verbose=False)

        psi2 = pe.periodic_extend(model1.psi, Mx, My)
        model2.set_dimensions(model1.Lx*Mx, model1.Ly*My, model1.Nx*Mx, model1.Ny*My, verbose=False)
        model2.set_psi(psi2, verbose=False)

        prompt.put(var2, model2)
        print()

               
    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 4:
            return _match_models(prompt, args[-1])
        return [] 


class CommandCrop(Command):
    def __init__(self):
        super().__init__('crop')


    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 6:
            raise CommandExecutionError('syntax: crop VAR1 ratio_x1 ratio_y1 ratio_x2 ratio_y2 VAR2')

        var1 = args[0]
        var2 = args[5]
        try:
            ratio_x1 = float(args[1])
            ratio_y1 = float(args[2])
            ratio_x2 = float(args[3])
            ratio_y2 = float(args[4])

            assert 0 <= ratio_x1 < ratio_x2 <= 1
            assert 0 <= ratio_y1 < ratio_y2 <= 1
            
        except ValueError as e:
            raise CommandExecutionError(e.args[-1])
        except AssertionError as e:
            raise CommandExecutionError('the ratios must satisfy the constraints ' +\
                                        '0 <= ratio_x1 < ratio_x2 <= 1 and ' +\
                                        '0 <= ratio_y1 < ratio_y2 <= 1')


        extra = args[6:]

        model1 = prompt.get(var1)
        model2 = model1.copy(verbose=False, copy_history=False)

        Nx2 = int(model1.Nx * (ratio_x2 - ratio_x1))
        Ny2 = int(model1.Ny * (ratio_y2 - ratio_y1))
        first_x = int(model1.Nx * ratio_x1)
        first_y = int(model1.Nx * ratio_y1)

        psi2 = model1.psi[first_x:first_x+Nx2, first_y:first_y+Ny2]

        model2.set_dimensions(model1.Lx*(ratio_x2-ratio_x1), model1.Ly*(ratio_y2-ratio_y1), Nx2, Ny2, verbose=False)
        model2.set_psi(psi2, verbose=False)
        prompt.put(var2, model2)
        print()

               
    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 6:
            return _match_models(prompt, args[-1])
        if 1 < len(args) < 6:
            return ['0.125 ', '0.25 ', '0.375 ', '0.5 ', '0.625 ', '0.75 ', '0.875 ', '1']
        return [] 


class CommandChangeRes(Command):
    def __init__(self):
        super().__init__('change-res')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 4:
            raise CommandExecutionError('syntax: change-res VAR1 NX NY VAR2')

        var1 = args[0]
        try:
            Nx = int(args[1])
            Ny = int(args[2])
        except ValueError as e:
            raise CommandExecutionError(e.args[-1])

        var2 = args[3]
        extra = args[4:]

        model1 = prompt.get(var1)
        model2 = model1.copy(verbose=False)


        psi2 = pe.change_resolution(model1.psi, Nx, Ny)

        model2.set_dimensions(model1.Lx, model1.Ly, Nx, Ny, verbose=False)
        model2.set_psi(psi2, verbose=False)

        prompt.put(var2, model2)
        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 4:
            return _match_models(prompt, args[-1])
        return [] 


class CommandCopy(Command):
    def __init__(self):
        super().__init__('copy')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 2:
            raise CommandExecutionError('syntax: copy VAR1 VAR2')

        var1 = args[0]
        var2 = args[1]

        model1 = prompt.get(var1)
        model2 = model1.copy(verbose=False)

        prompt.put(var2, model2)
        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 2:
            return _match_models(prompt, args[-1])
        return [] 


class CommandDelete(Command):
    def __init__(self):
        super().__init__('del')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 1:
            raise CommandExecutionError('syntax: delete VAR')

        var = args[0]

        if not var in prompt.memory_models:
            raise CommandExecutionError(f'{var} does not exist in memory')

        try:
            del prompt.memory_models[var]
        except Exception:
            raise CommandExecutionError(f'could not delete {var} from memory')

        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1:
            return _match_models(prompt, args[-1])
        return [] 


class CommandMove(Command):
    def __init__(self):
        super().__init__('move')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 2:
            raise CommandExecutionError('syntax: move VAR1 VAR2')

        var1 = args[0]
        var2 = args[1]

        if not var1 in prompt.memory_models:
            raise CommandExecutionError(f'{var1} does not exist in memory')

        prompt.put(var2, prompt.get(var1))
        del prompt.memory_models[var1]

        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 2:
            return _match_models(prompt, args[-1])
        return [] 


class CommandSetParam(Command):
    def __init__(self):
        super().__init__('set-param')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 4:
            raise CommandExecutionError('syntax: set-param VAR1 MU EPS VAR2')

        var1 = args[0]
        try:
            mu2 = float(args[1])
            eps2 = float(args[2])
        except ValueError as e:
            raise CommandExecutionError(e.args[-1])
        var2 = args[3]

        extra = args[4:]

        model1 = prompt.get(var1)
        model2 = model1.copy(verbose=False)

        model2.set_params(mu2, eps2, verbose=False)

        prompt.put(var2, model2)
        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 4:
            return _match_models(prompt, args[-1])
        return [] 


class CommandSetSize(Command):
    def __init__(self):
        super().__init__('set-size')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 4:
            raise CommandExecutionError('syntax: set-size VAR1 LX LY VAR2')

        var1 = args[0]
        try:
            if args[1].endswith('ux'):
                Lx = float(args[1][:-2]) * 4*np.pi
            elif args[1].endswith('uy'):
                Lx = float(args[1][:-2]) * 4*np.pi / np.sqrt(3)
            else:
                Lx = float(args[1])

            if args[2].endswith('ux'):
                Ly = float(args[2][:-2]) * 4*np.pi
            elif args[2].endswith('uy'):
                Ly = float(args[2][:-2]) * 4*np.pi / np.sqrt(3)
            else:
                Ly = float(args[2])

        except ValueError as e:
            raise CommandExecutionError(e.args[-1])

        var2 = args[3]

        model1 = prompt.get(var1)
        model2 = model1.copy(verbose=False)
        model2.resize(Lx, Ly, verbose=False)

        prompt.put(var2, model2)
        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 4:
            return _match_models(prompt, args[-1])
        return []


class CommandSetMinimizer(Command):
    def __init__(self):
        super().__init__('set-minimizer')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 4:
            raise CommandExecutionError('syntax: set-minimizer VAR1 minimizer dt VAR2')

        var1 = args[0]
        try:
            minimizer = args[1]
            dt = float(args[2])
            assert minimizer in ['mu', 'nonlocal']
        except ValueError as e:
            raise CommandExecutionError(e.args[-1])
        except AssertionError as e:
            raise CommandExecutionError('minimizer must be either \'mu\' or \'nonlocal\'')


        var2 = args[3]
        extra = args[4:]

        model1 = prompt.get(var1)
        model2 = model1.copy(verbose=False)
        model2.minimizer = minimizer
        model2.dt = dt

        prompt.put(var2, model2)
        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 4:
            return _match_models(prompt, args[-1])
        if len(args) == 2:
            return ['mu', 'nonlocal']
        return [] 


class CommandInterface(Command):
    def __init__(self):
        super().__init__('interface')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 4:
            raise CommandExecutionError('syntax: interface VAR1 VAR2 WIDTH OUTPUT')

        var_sol = args[0]
        var_liq = args[1]
        var_out = args[3]
        try:
            width = float(args[2])
        except ValueError as e:
            raise CommandExecutionError(e.args[-1])

        extra = args[4:]

        model_sol = prompt.get(var_sol)
        model_liq = prompt.get(var_liq)

        try:
            assert abs(model_sol.Lx - model_liq.Lx) / model_sol.Lx < 1e-6
            assert abs(model_sol.Ly - model_liq.Ly) / model_sol.Ly < 1e-6
            assert model_sol.Nx == model_liq.Nx
            assert model_sol.Ny == model_liq.Ny

        except AssertionError as e:
            raise CommandExecutionError(f'models {var_sol} and {var_liq} do not have the same dimensions')

        model_out = model_sol.copy(verbose=False, copy_history=False)

        X, Y = model_sol.X, model_sol.Y
        xa = model_sol.Lx / 4
        xb = model_sol.Lx / 4 * 3

        bump = (1+np.tanh((X-xa)/width))/2 * (1+np.tanh((-X+xb)/width))/2
        psi_out = model_sol.psi * bump + model_liq.psi * (1-bump)
        model_out.set_psi(psi_out, verbose=False)

        prompt.put(var_out, model_out)
        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 2 or len(args) == 4:
            return _match_models(prompt, args[-1])
        return [] 


class CommandLiquefy(Command):
    def __init__(self):
        super().__init__('liquefy')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 2:
            raise CommandExecutionError('syntax: liquefy VAR1 VAR2 [--density psi0]')

        var = args[0]
        var_out = args[1]
        extra = args[2:]
        
        model = prompt.get(var)
        param = {'density': model.calc_N_tot()/model.Volume}

        i = 0
        while i < len(extra):
            if extra[i] == '--density':
                try:
                    assert len(extra) > i+1
                except AssertionError as e:
                    raise CommandExecutionError('please specify a density after --density')
                param['density'] = float(extra[i+1])
                i += 1
            else:
                raise CommandExecutionError(f'unknown option: {extra[i]}')

            i += 1


        model_out = model.copy(verbose=False)

        model_out.set_psi(param['density'] + 0*model.X, verbose=False)

        prompt.put(var_out, model_out)
        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 2:
            return _match_models(prompt, args[-1])
        if len(args) == 3:
            return ['--density ']
        return [] 


class CommandInsert(Command):
    def __init__(self):
        super().__init__('insert')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 3:
            raise CommandExecutionError(
                'syntax: insert SMALL LARGE VAR [--position left|right|top|bottom|center] [--offset X Y]')

        var_s = args[0]
        var_l = args[1]
        var_out = args[2]
        extra = args[3:]
        
        model_s = prompt.get(var_s)
        model_l = prompt.get(var_l)
        param = {'position': 'center', 'offset': (0,0), 'width': 0}

        if not (model_s.Lx <= model_l.Lx and model_s.Ly <= model_l.Ly):
            raise CommandExecutionError('the first model must have smaller dimensions than the second one')


        # parse extra arguments
        i = 0
        while i < len(extra):
            if extra[i] == '--position':
                if len(extra) <= i+1:
                    raise CommandExecutionError('please specify a position after --position')
                if not extra[i+1] in ['left', 'right', 'top' 'bottom', 'center']:
                    raise CommandExecutionError(f'invalid position: {extra[i+1]}')
                param['position'] = extra[i+1]
                i += 1

            elif extra[i] == '--offset':
                if len(extra) <= i+2:
                    raise CommandExecutionError('please specify x and y offsets after --offset')
                try:
                    param['offset'] = (float(extra[i+1]), float(extra[i+2]))
                except ValueError as e:
                    raise CommandExecutionError(f'invalid offset: {extra[i+1]} {extra[i+2]}')
                i += 2

            else:
                raise CommandExecutionError(f'unkonw option: {extra[i]}')

            i += 1
        
        psi_small = model_s.psi.copy()
        psi_large = model_l.psi.copy()

        box_Nx, box_Ny = int(model_s.Lx / model_l.Lx * model_l.Nx), int(model_s.Ly / model_l.Ly * model_l.Ny)
        
        psi_small_modified = pe.change_resolution(psi_small, box_Nx, box_Ny)
       
        box_X, box_Y = int(model_l.Nx/2 - box_Nx/2), int(model_l.Ny/2 - box_Ny/2)
        if param['position'] == 'left':
            box_X = 0
            
        if param['position'] == 'right':
            box_X = model_l.Nx - box_Nx

        if param['position'] == 'top':
            box_Y = 0
            
        if param['position'] == 'bottom':
            box_Y = model_l.Ny - box_Ny

        psi_large[box_X:box_X+box_Nx, box_Y:box_Y+box_Ny] = psi_small_modified
        

        model_out = model_l.copy(verbose=False, copy_history=False)
        model_out.set_psi(psi_large, verbose=False)

        prompt.put(var_out, model_out)
        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 2 or len(args) == 3:
            return _match_models(prompt, args[-1])
        if len(args) >= 4:
            if args[-2] == '--position':
                return ['top', 'bottom', 'left', 'right', 'center']
            if args[-2] == '--offset':
                return []
            return ['--position', '--offset']
        return [] 

class CommandFlip(Command):
    def __init__(self):
        super().__init__('flip')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 3:
            raise CommandExecutionError(
                'syntax: flip VAR [X|Y] VAROUT')

        var = args[0]
        axis = args[1]
        var_out = args[2]
        extra = args[3:]
        
        model = prompt.get(var)

        model_out = model.copy(verbose=False)

        if axis == 'Y':
            model_out.set_psi(model.psi[::-1,:], verbose=False)
        elif axis == 'X':
            model_out.set_psi(model.psi[:,::-1], verbose=False)
        else:
            raise CommandExecutionError('axis must be either X or Y')

        prompt.put(var_out, model_out)
        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 3:
            return _match_models(prompt, args[-1])
        if len(args) >= 2:
            return ['X', 'Y']
        return []

class CommandRotate(Command):
    def __init__(self):
        super().__init__('rotate')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 3:
            raise CommandExecutionError(
                'syntax: rotate VAR [90|180|270] VAROUT')

        var = args[0]
        angle = args[1]
        var_out = args[2]
        extra = args[3:]
        
        model = prompt.get(var)
        model_out = model.copy(verbose=False)

        if angle == '90':
            model_out.set_dimensions(model.Ly, model.Lx, model.Ny, model.Nx, verbose=False)
            model_out.set_psi(np.transpose(model.psi)[:,::-1], verbose=False)
        elif angle == '180':
            model_out.set_psi(model.psi[::-1,::-1], verbose=False)
        elif angle == '270':
            model_out.set_dimensions(model.Ly, model.Lx, model.Ny, model.Nx, verbose=False)
            model_out.set_psi(np.transpose(model.psi)[::-1,:], verbose=False)
        else:
            raise CommandExecutionError('axis must be 90, 180, or 270')

        prompt.put(var_out, model_out)
        print()

    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 3:
            return _match_models(prompt, args[-1])
        if len(args) == 2:
            return ['X', 'Y']
        return []



class CommandConcat(Command):
    def __init__(self):
        super().__init__('concat')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 4:
            raise CommandExecutionError(
                'syntax: concat VAR1 VAR2 [horizontal|vertical] VAROUT')

        var1 = args[0]
        var2 = args[1]
        concat_direction = args[2]
        var_out = args[3]
        extra = args[4:]
        
        model1 = prompt.get(var1)
        model2 = prompt.get(var2)

        model_out = model1.copy(verbose=False, copy_history=False)

        if concat_direction == 'vertical':
            if model1.Nx != model2.Nx or model1.Lx != model2.Lx:
                raise CommandExecutionError(f'models {var1} and {var2} do not have the same width')
            model_out.set_dimensions(model1.Lx, model1.Ly+model2.Ly, model1.Nx, model1.Ny+model2.Ny, verbose=False)
            psi_out = np.zeros(shape=(model1.Nx, model1.Ny+model2.Ny))
            psi_out[:,:model1.Ny] = model1.psi[:,:]
            psi_out[:,model1.Ny:] = model2.psi[:,:]

            model_out.set_psi(psi_out)

        elif concat_direction == 'horizontal':
            if model1.Ny != model2.Ny or model1.Ly != model2.Ly:
                raise CommandExecutionError(f'models {var1} and {var2} do not have the same height')
            model_out.set_dimensions(model1.Lx+model2.Lx, model1.Ly, model1.Nx+model2.Nx, model1.Ny, verbose=False)
            psi_out = np.zeros(shape=(model1.Nx+model2.Nx, model1.Ny)) 
            psi_out[:model1.Nx,:] = model1.psi[:,:]
            psi_out[model1.Nx:,] = model2.psi[:,:]
            model_out.set_psi(psi_out)

        else:
            raise CommandExecutionError(f'conact direction must be either horizontal or vertical')

        prompt.put(var_out, model_out)
        print()

    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1 or len(args) == 2 or len(args) == 4:
            return _match_models(prompt, args[-1])
        if len(args) == 3:
            return ['horizontal', 'vertical']
        return []


class CommandEvolve(Command):
    def __init__(self):
        super().__init__('evolve')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 1:
            raise CommandExecutionError('syntax: evolve VAR [options]')

        var = args[0]
        extra = args[1:]

        params = {'lazy': 1, 'display-precision': 5}
        
        i = 0
        try:
            while i < len(extra):
                if extra[i] == '--lazy-plot':
                    assert i+1 < len(extra)
                    params['lazy'] = int(extra[i+1])
                    i += 1
                if extra[i] == '--display-precision':
                    assert i+1 < len(extra)
                    params['display-precision'] = int(extra[i+1])
                i += 1
        except AssertionError:
            prompt.output('error occured while parsing arguments:', level=1)
            prompt.output(f'{extra}', level=1)

        model = prompt.get(var)


        lock, thread = None, None
        if model.minimizer == 'mu':
            lock, thread = model.run_background(model.minimize_mu, (model.dt,), 
                                                {'display_precision': params['display-precision']})
        if model.minimizer == 'nonlocal':
            lock, thread = model.run_background(model.minimize_nonlocal_conserved, (model.dt,), 
                                                {'display_precision': params['display-precision']})

        while True:
            try:
                time.sleep(1)

            except KeyboardInterrupt:
                with lock:
                    print()
                    prompt.output('------- pfc_im minimizer interrupted --------')
                    prompt.output('0: plot profile')
                    prompt.output('3: continue')
                    prompt.output('8: end minimization')
                    print('---------------------------------------------')

                    resp = input('enter action: ')

                    if resp in ['0']:
                        model.plot(lazy_factor=params['lazy'])

                    if resp in ['8']:
                        model.stop_minimization()
                        print()
                        return

    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1:
            return _match_models(prompt, args[-1])
        if len(args) >= 2:
            if args[-2] == '--lazy-plot' or args[-2] == '--display-precision':
                return [] 
            return ['--lazy-plot ', '--display-precision ']
        return [] 


class CommandHistory(Command):
    def __init__(self):
        super().__init__('history')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 1:
            raise CommandExecutionError('syntax: history VAR')

        var = args[0]
        extra = args[1:]

        model = prompt.get(var)
        ax1 = plt.subplot(211, title='history - grand potential density', xlabel='t', ylabel=r'$\omega$')
        ax2 = plt.subplot(212, title='history - mean density', xlabel='t', ylabel=r'$\bar\psi$')

        try:
            ax1.plot(model.history['t'], model.history['omega'], color='red')
            ax2.plot(model.history['t'], model.history['psi0'], color='black')
        except KeyError as e:
            raise CommandExecutionError(f'history file of {var} is corrupted, consider reseting history')

        plt.tight_layout()
        plt.show()

        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1:
            return _match_models(prompt, args[-1])
        return [] 


class CommandResetHistory(Command):
    def __init__(self):
        super().__init__('reset-history')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) < 1:
            raise CommandExecutionError('syntax: reset-history VAR')

        var = args[0]
        extra = args[1:]

        model = prompt.get(var)
        model.new_history()

        print()


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        if len(args) == 1:
            return _match_models(prompt, args[-1])
        return [] 


class CommandIPython(Command):
    ipython_session = '.pipe_ipython_session'
    ipython_startup = '.pipe_ipython_startup.py'
    def __init__(self):
        super().__init__('ipython3')
        os.remove(self.ipython_session)
        open(self.ipython_session, 'a').close()

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        candidates = [f'ipython3 --no-banner -i {self.ipython_startup}', 'python3', 'ipython', 'python']
        terminal_width = 80
        try:
            terminal_width = os.get_terminal_size().columns
        except:
            pass


        #prompt.output(f'launching {candidates[0]} ...')
        print(cmn.bcolors.OKGREEN + '┌' + '─' * (terminal_width//4*3) + '┐' + cmn.bcolors.ENDC)
        os.system(candidates[0])

        print(cmn.bcolors.OKGREEN + '└' + '─' * (terminal_width//4*3) + '┘' + cmn.bcolors.ENDC)
        #prompt.output(f'terminating {candidates[0]} ...')
        print()
        return


         


    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        return [] 


class CommandExit(Command):
    def __init__(self):
        super().__init__('exit')

    @overrides(Command)
    def execute(self, prompt: ProfileEditorPrompt, args: list):
        prompt.running = False
                
    @overrides(Command)
    def complete(self, prompt: ProfileEditorPrompt, args: list):
        return [] 



def _match_models(prompt, arg):
    return [f'{mod}' for mod in prompt.memory_models.keys() if mod.startswith(arg)]


